"""
Tools for interacting with the DOM inside a browser.
"""

from copy import copy
from collections import Sequence
from itertools import islice
from selenium.common.exceptions import WebDriverException

from bok_choy.promise import Promise


# Mapping of query type to Selenium webdriver query method names
QUERY_TYPES = {
    'css': 'find_elements_by_css_selector',
    'xpath': 'find_elements_by_xpath',
}


def no_error(func):
    """
    Decorator to create a `Promise` check function that is satisfied
    only when `func` executes without a Selenium error.

    This protects against many common test failures due to timing issues.
    For example, accessing an element after it has been modified by JavaScript
    ordinarily results in a `StaleElementException`.  Methods decorated
    with `no_error` will simply retry if that happens, which makes tests
    more robust.

    Args:
        func (callable): The function to execute, with retries if an error occurs.

    Returns:
        Decorated function
    """
    def _inner(*args, **kwargs):  # pylint: disable=missing-docstring
        try:
            return_val = func(*args, **kwargs)
        except WebDriverException:
            return (False, None)
        else:
            return (True, return_val)

    return _inner


class Query(Sequence):
    """
    General mechanism for selecting and transforming values.
    """

    def __init__(self, seed_fn, desc=None):
        """
        Configure the `Query`.

        Args:
            seed_fn (callable): Callable with no arguments that produces a list of values.

        Keyword Args:
            desc (str): A description of the query, used in log messages.
                If not provided, defaults to the name of the seed function.

        Returns:
            Query
        """
        if desc is None:
            desc = u'Query({})'.format(getattr(seed_fn, '__name__', ''))

        self.seed_fn = seed_fn
        self.transforms = []
        self.desc_stack = []
        self.desc = desc

    def replace(self, **kwargs):
        """
        Return a copy of this `Query`, but with attributes specified
        as keyword arguments replaced by the keyword values.

        Keyword Args:
            Attributes/values to replace in the copy.

        Returns:
             A copy of the query that has its attributes updated with the specified values.

        Raises:
            TypeError: The `Query` does not have the specified attribute.
        """
        clone = copy(self)

        clone.transforms = list(clone.transforms)
        for key, value in kwargs.items():
            if not hasattr(clone, key):
                raise TypeError('replace() got an unexpected keyword argument {!r}'.format(key))

            setattr(clone, key, value)
        return clone

    def transform(self, transform, desc=None):
        """
        Create a copy of this query, transformed by `transform`.

        Args:
            transform (callable): Callable that takes an iterable of values and
                returns an iterable of transformed values.

        Keyword Args:
            desc (str): A description of the transform, to use in log messages.
                Defaults to the name of the `transform` function.

        Returns:
            Query
        """
        if desc is None:
            desc = u'transform({})'.format(getattr(transform, '__name__', ''))

        return self.replace(
            transforms=self.transforms + [transform],
            desc_stack=self.desc_stack + [desc]
        )

    def map(self, map_fn, desc=None):
        """
        Return a copy of this query, with the values mapped through `map_fn`.

        Args:
            map_fn (callable): A callable that takes a single argument and returns a new value.

        Keyword Args:
            desc (str): A description of the mapping transform, for use in log message.
                Defaults to the name of the map function.

        Returns:
            Query
        """
        if desc is None:
            desc = getattr(map_fn, '__name__', '')
        desc = u'map({})'.format(desc)

        return self.transform(lambda xs: (map_fn(x) for x in xs), desc=desc)

    def filter(self, filter_fn=None, desc=None, **kwargs):
        """
        Return a copy of this query, with some values removed.

        Example usages:

        .. code:: python

            # Returns a query that matches even numbers
            q.filter(filter_fn=lambda x: x % 2)

            # Returns a query that matches elements with el.description == "foo"
            q.filter(description="foo")

        Keyword Args:
            filter_fn (callable): If specified, a function that accepts one argument (the element)
                    and returns a boolean indicating whether to include that element in the results.

            kwargs: Specify attribute values that an element must have to be included in the results.

            desc (str): A description of the filter, for use in log messages.
                Defaults to the name of the filter function or attribute.

        Raises:
            TypeError: neither or both of `filter_fn` and `kwargs` are provided.
        """
        if filter_fn is not None and kwargs:
            raise TypeError('Must supply either a filter_fn or attribute filter parameters to filter(), but not both.')
        if filter_fn is None and not kwargs:
            raise TypeError('Must supply one of filter_fn or one or more attribute filter parameters to filter().')

        if desc is None:
            if filter_fn is not None:
                desc = getattr(filter_fn, '__name__', '')
            elif kwargs:
                desc = u", ".join([u"{}={!r}".format(key, value) for key, value in kwargs.items()])
        desc = u"filter({})".format(desc)

        if kwargs:
            def filter_fn(elem):  # pylint: disable=function-redefined, missing-docstring
                return all(
                    getattr(elem, filter_key) == filter_value
                    for filter_key, filter_value
                    in kwargs.items()
                )

        return self.transform(lambda xs: (x for x in xs if filter_fn(x)), desc=desc)

    def _execute(self):
        """
        Run the query, generating data from the `seed_fn` and performing transforms on the results.
        """
        data = self.seed_fn()
        for transform in self.transforms:
            data = transform(data)
        return list(data)

    def execute(self, try_limit=5, try_interval=0.5, timeout=30):
        """
        Execute this query, retrying based on the supplied parameters.

        Keyword Args:
            try_limit (int): The number of times to retry the query.
            try_interval (float): The number of seconds to wait between each try (float).
            timeout (float): The maximum number of seconds to spend retrying (float).

        Returns:
            The transformed results of the query.

        Raises:
            BrokenPromise: The query did not execute without a Selenium error after one or more attempts.
        """
        return Promise(
            no_error(self._execute),
            u"Executing {!r}".format(self),
            try_limit=try_limit,
            try_interval=try_interval,
            timeout=timeout,
        ).fulfill()

    @property
    def results(self):
        """
        A list of the results of the query, which are cached.
        If you call `results` multiple times on the same query, you will always get the same results.
        Use `reset()` to clear the cache and re-run the query.

        Returns:
            The results from executing the query.
        """
        return self.execute()

    def __getitem__(self, key):
        return self.results[key]

    def __len__(self):
        return len(self.results)

    def is_present(self):
        """
        Check whether the query returns any results.

        Returns:
            Boolean indicating whether the query contains any results.
        """
        return bool(self.results)

    present = property(is_present)

    @property
    def first(self):
        """
        Return a Query that selects only the first element of this Query.
        If no elements are available, returns a query with no results.

        Example usage:

        .. code:: python

            >> q = Query(lambda: range(5))
            >> q.first.results
            [0]

        Returns:
            Query
        """
        def _transform(xs):  # pylint: disable=missing-docstring, invalid-name
            try:
                return [iter(xs).next()]
            except StopIteration:
                return []

        return self.transform(_transform, 'first')

    def nth(self, index):
        """
        Return a query that selects the element at `index` (starts from 0).
        If no elements are available, returns a query with no results.

        Example usage:

        .. code:: python

            >> q = Query(lambda: range(5))
            >> q.nth(2).results
            [2]

        Args:
            index (int): The index of the element to select (starts from 0)

        Returns:
            Query
        """
        def _transform(xs):  # pylint: disable=missing-docstring, invalid-name
            try:
                return [next(islice(iter(xs), index, None))]

            # Gracefully handle (a) running out of elements, and (b) negative indices
            except (StopIteration, ValueError):
                return []

        return self.transform(_transform, 'nth')

    def __repr__(self):
        return u".".join([self.desc] + self.desc_stack)


class BrowserQuery(Query):
    """
    A Query that operates on a browser.
    """
    def __init__(self, browser, **kwargs):
        """
        Generate a query over a browser.

        Args:
            browser (selenium.webdriver): A Selenium-controlled browser.

        Keyword Args:
            css (str): A CSS selector.
            xpath (str): An XPath selector.

        Returns:
            BrowserQuery

        Raises:
            TypeError: The query must be passed either a CSS or XPath selector, but not both.
        """
        if len(kwargs) > 1:
            raise TypeError('BrowserQuery() takes at most 1 keyword argument.')

        if not kwargs:
            raise TypeError('Must pass a query keyword argument to BrowserQuery().')

        query_name, query_value = kwargs.items()[0]

        if query_name not in QUERY_TYPES:
            raise TypeError('{} is not a supported query type for BrowserQuery()'.format(query_name))

        def query_fn():  # pylint: disable=missing-docstring
            return getattr(browser, QUERY_TYPES[query_name])(query_value)

        super(BrowserQuery, self).__init__(
            query_fn,
            desc=u"BrowserQuery({}={!r})".format(query_name, query_value),
        )
        self.browser = browser

    def attrs(self, attribute_name):
        """
        Retrieve HTML attribute values from the elements matched by the query.

        Example usage:

        .. code:: python

            # Assume that the query matches html elements:
            # <div class="foo"> and <div class="bar">
            >> q.attrs('class')
            ['foo', 'bar']

        Args:
            attribute_name (str): The name of the attribute values to retrieve.

        Returns:
            A list of attribute values for `attribute_name`.
        """
        desc = 'attrs({!r})'.format(attribute_name)
        return self.map(lambda el: el.get_attribute(attribute_name), desc).results

    @property
    def text(self):
        """
        Retrieve text from each matched element.

        Example usage:

        .. code:: python

            # Assume that the query matches html elements:
            # <div>Foo</div> and <div>Bar</div>
            >> q.text
            ['Foo', 'Bar']

        Returns:
            The text of each element matched by the query.
        """
        return self.map(lambda el: el.text, 'text').results

    @property
    def html(self):
        """
        Retrieve the inner HTML of each element matched by the query.

        Example usage:

        .. code:: python

            # Assume that the query matches html elements:
            # <div><span>Foo</span></div> and <div>Bar</div>
            >> q.html
            ['<span>Foo</span>', 'Bar']

        Returns:
            The inner HTML for each element matched by the query.
        """
        return self.map(lambda el: el.get_attribute('innerHTML'), 'html').results

    @property
    def selected(self):
        """
        Check whether all the matched elements are selected.

        Returns:
            bool
        """
        query_results = self.map(lambda el: el.is_selected(), 'selected').results
        if query_results:
            return all(query_results)
        else:
            return False

    @property
    def visible(self):
        """
        Check whether all matched elements are visible.

        Returns:
            bool
        """
        query_results = self.map(lambda el: el.is_displayed(), 'visible').results
        if query_results:
            return all(query_results)
        else:
            return False

    @property
    def invisible(self):
        """
        Check whether all matched elements are present, but not visible.

        Returns:
            bool
        """
        return self.present and not self.visible

    def is_focused(self):
        """
        Checks that *at least one* matched element is focused. More
        specifically, it checks whether the element is document.activeElement.
        If no matching element is focused, this returns `False`.

        Returns:
            bool
        """
        active_el = self.browser.execute_script("return document.activeElement")
        query_results = self.map(lambda el: el == active_el, 'focused').results

        if query_results:
            return any(query_results)
        else:
            return False

    focused = property(is_focused)

    def click(self):
        """
        Click each matched element.

        Example usage:

        .. code:: python

            # Click the first element matched by the query
            q.first.click()

        Returns:
            None
        """
        self.map(lambda el: el.click(), 'click()').execute()

    def fill(self, text):
        """
        Set the text value of each matched element to `text`.

        Example usage:

        .. code:: python

            # Set the text of the first element matched by the query to "Foo"
            q.first.fill('Foo')

        Args:
            text (str): The text used to fill the element (usually a text field or text area).

        Returns:
            None
        """
        def _fill(elem):  # pylint: disable=missing-docstring
            elem.clear()
            elem.send_keys(text)

        self.map(_fill, 'fill({!r})'.format(text)).execute()



"""
Helpers for dealing with JavaScript synchronization issues.
"""
import functools
import json
from textwrap import dedent
from selenium.common.exceptions import TimeoutException, WebDriverException
from .promise import EmptyPromise


def js_defined(*js_vars):
    """
    Class decorator that ensures JavaScript variables are defined in the browser.

    This adds a `wait_for_js` method to the class, which will
    block until all the expected JavaScript variables are defined.

    Args:
        js_vars (list of str): List of JavaScript variable names to wait for.

    Returns:
        Decorated class
    """
    return _decorator('_js_vars', js_vars)


def requirejs(*modules):
    """
    Class decorator that ensures RequireJS modules are loaded in the browser.

    This adds a `wait_for_js` method to the class, which will
    block until all the expected RequireJS modules are loaded.

    Args:
        modules (list of str) List of RequireJS module names to wait for.

    Returns:
        Decorated class
    """
    return _decorator('_requirejs_deps', modules)


def wait_for_js(function):
    """
    Method decorator that waits for JavaScript dependencies before executing `function`.
    If the function is not a method, the decorator has no effect.

    Args:
        function (callable): Method to decorate.

    Returns:
        Decorated method
    """

    @functools.wraps(function)
    def wrapper(*args, **kwargs):  # pylint: disable=missing-docstring

        # If not a method, then just call the function
        if len(args) < 1:
            return function(*args, **kwargs)

        # Otherwise, retrieve `self` as the first arg
        else:
            self = args[0]

            # If the class has been decorated by one of the
            # JavaScript dependency decorators, it should have
            # a `wait_for_js` method
            if hasattr(self, 'wait_for_js'):
                self.wait_for_js()

            # Call the function
            return function(*args, **kwargs)

    return wrapper


def _decorator(store_name, store_values):
    """
    Return a class decorator that:

    1) Defines a new class method, `wait_for_js`
    2) Defines a new class list variable, `store_name` and adds
        `store_values` to the list.
    """
    def decorator(clz):  # pylint: disable=missing-docstring

        # Add a `wait_for_js` method to the class
        if not hasattr(clz, 'wait_for_js'):
            setattr(clz, 'wait_for_js', _wait_for_js)  # pylint: disable= literal-used-as-attribute

        # Store the RequireJS module names in the class
        if not hasattr(clz, store_name):
            setattr(clz, store_name, set())

        getattr(clz, store_name).update(store_values)
        return clz

    return decorator


def _wait_for_js(self):
    """
    Class method added by the decorators to allow
    decorated classes to manually re-check JavaScript
    dependencies.

    Expect that `self` is a class that:
    1) Has been decorated with either `js_defined` or `requirejs`
    2) Has a `browser` property

    If either (1) or (2) is not satisfied, then do nothing.
    """

    # No Selenium browser available, so return without doing anything
    if not hasattr(self, 'browser'):
        return

    # pylint: disable=protected-access
    # Wait for JavaScript variables to be defined
    if hasattr(self, '_js_vars') and self._js_vars:
        EmptyPromise(
            lambda: _are_js_vars_defined(self.browser, self._js_vars),
            "JavaScript variables defined: {0}".format(", ".join(self._js_vars))
        ).fulfill()

    # Wait for RequireJS dependencies to load
    if hasattr(self, '_requirejs_deps') and self._requirejs_deps:
        EmptyPromise(
            lambda: _are_requirejs_deps_loaded(self.browser, self._requirejs_deps),
            "RequireJS dependencies loaded: {0}".format(", ".join(self._requirejs_deps)),
            try_limit=5
        ).fulfill()


def _are_js_vars_defined(browser, js_vars):
    """
    Return a boolean indicating whether all the JavaScript
    variables `js_vars` are defined on the current page.

    `browser` is a Selenium webdriver instance.
    """
    # This script will evaluate to True iff all of
    # the required vars are defined.
    script = " && ".join([
        "!(typeof {0} === 'undefined')".format(var)
        for var in js_vars
    ])

    try:
        return browser.execute_script("return {}".format(script))
    except WebDriverException as exc:
        if "is not defined" in exc.msg or "is undefined" in exc.msg:
            return False
        else:
            raise


def _are_requirejs_deps_loaded(browser, deps):
    """
    Return a boolean indicating whether all the RequireJS
    dependencies `deps` have loaded on the current page.

    `browser` is a WebDriver instance.
    """

    # This is a little complicated
    #
    # We're going to use `execute_async_script` to give control to
    # the browser.  The browser indicates that it wants to return
    # control to us by calling `callback`, which is the last item
    # in the global `arguments` array.
    #
    # We install a RequireJS module with the dependencies we want
    # to ensure are loaded.  When our module loads, we return
    # control to the test suite.
    script = dedent("""
        // Retrieve the callback function used to return control to the test suite
        var callback = arguments[arguments.length - 1];

        // If RequireJS isn't defined, then return immediately
        if (!window.require) {{
            callback("RequireJS not defined");
        }}

        // Otherwise, install a RequireJS module that depends on the modules
        // we're waiting for.
        else {{

            // Catch errors reported by RequireJS
            requirejs.onError = callback;

            // Install our module
            require({deps}, function() {{
                callback('Success');
            }});
        }}
    """).format(deps=json.dumps(list(deps)))

    # Set a timeout to ensure we get control back
    browser.set_script_timeout(30)

    # Give control to the browser
    # `result` will be the argument passed to the callback function
    try:
        result = browser.execute_async_script(script)
        return result == 'Success'

    except TimeoutException:
        return False



"""
Interface for using the axe-core ruleset.
See: https://github.com/dequelabs/axe-core
"""
import json
import os

from textwrap import dedent, fill

from .a11y_audit import A11yAudit, A11yAuditConfig, AccessibilityError, A11yAuditConfigError
from ..promise import Promise


CUR_DIR = os.path.dirname(os.path.abspath(__file__))


class AxeCoreAuditConfig(A11yAuditConfig):
    """
    The `AxeCoreAuditConfig` object defines the options available when
    running an `AxeCoreAudit`.
    """
    def __init__(self, *args, **kwargs):
        super(AxeCoreAuditConfig, self).__init__(*args, **kwargs)
        self.rules, self.context = None, None
        self.custom_rules = "customRules={}"
        self.rules_file = os.path.join(
            os.path.split(CUR_DIR)[0],
            'vendor/axe-core/axe.min.js'
        )

        self.set_rules({})
        self.set_scope()
        self.customize_ruleset()

    def set_rules(self, rules):
        """
        Set rules to ignore XOR limit to when checking for accessibility
        errors on the page.

        Args:

            rules: a dictionary one of the following formats.
                If you want to run all of the rules except for some::

                    {"ignore": []}

                If you want to run only a specific set of rules::

                    {"apply": []}

                If you want to run only rules of a specific standard::

                    {"tags": []}

        Examples:

            To run only "bad-link" and "color-contrast" rules::

                page.a11y_audit.config.set_rules({
                    "apply": ["bad-link", "color-contrast"],
                })

            To run all rules except for "bad-link" and "color-contrast"::

                page.a11y_audit.config.set_rules({
                    "ignore": ["bad-link", "color-contrast"],
                })

            To run only WCAG 2.0 Level A rules::

                page.a11y_audit.config.set_rules({
                    "tags": ["wcag2a"],
                })

            To run all rules:
                page.a11y_audit.config.set_rules({})

        Related documentation:

            * https://github.com/dequelabs/axe-core/blob/master/doc/API.md#options-parameter-examples
            * https://github.com/dequelabs/axe-core/doc/rule-descriptions.md
        """
        options = {}
        if rules:
            if rules.get("ignore"):
                options["rules"] = {}
                for rule in rules.get("ignore"):
                    options["rules"][rule] = {"enabled": False}
            elif rules.get("apply"):
                options["runOnly"] = {
                    "type": "rule",
                    "values": rules.get("apply"),
                }
            elif rules.get("tags"):
                options["runOnly"] = {
                    "type": "tag",
                    "values": rules.get("tags"),
                }
        self.rules = json.dumps(options)

    def set_scope(self, include=None, exclude=None):
        """
        Sets `scope` (refered to as `context` in ruleset documentation), which
        defines the elements on a page to include or exclude in the audit. If
        neither `include` nor `exclude` are passed, the entire document will
        be included.

        Args:

            include (optional): a list of css selectors for elements that
            should be included in the audit. By, default, the entire document
            is included.
            exclude (optional): a list of css selectors for elements that should not
            be included in the audit.

        Examples:

            To include all items in `#main-content` except `#some-special-elm`::

                page.a11y_audit.config.set_scope(
                    exclude=["#some-special-elm"],
                    include=["#main-content"]
                )

            To include all items in the document except `#some-special-elm`::

                page.a11y_audit.config.set_scope(
                    exclude=["#some-special-elm"],
                )

            To include only children of `#some-special-elm`::

                page.a11y_audit.config.set_scope(
                    include=["#some-special-elm"],
                )

        Context documentation:

            https://github.com/dequelabs/axe-core/blob/master/doc/API.md#a-context-parameter

            Note that this implementation only supports css selectors. It does
            not accept nodes as described in the above documentation resource.
        """
        context = {}

        if exclude:
            context["exclude"] = [[selector] for selector in exclude]

        if include:
            context["include"] = [[selector] for selector in include]

        self.context = json.dumps(context) if context else 'document'

    def customize_ruleset(self, custom_ruleset_file=None):
        """
        Updates the ruleset to include a set of custom rules. These rules will
        be _added_ to the existing ruleset or replace the existing rule with
        the same ID.

        Args:

            custom_ruleset_file (optional): The filepath to the custom rules.
                Defaults to `None`. If `custom_ruleset_file` isn't passed, the
                environment variable `BOKCHOY_A11Y_CUSTOM_RULES_FILE` will be
                checked. If a filepath isn't specified by either of these
                methods, the ruleset will not be updated.

        Raises:

            `IOError` if the specified file does not exist.

        Examples:

            To include the rules defined in `axe-core-custom-rules.js`::

                page.a11y_audit.config.customize_ruleset(
                    "axe-core-custom-rules.js"
                )

            Alternatively, use the environment variable `BOKCHOY_A11Y_CUSTOM_RULES_FILE`
            to specify the path to the file containing the custom rules.

        Documentation for how to write rules:

            https://github.com/dequelabs/axe-core/blob/master/doc/developer-guide.md

        An example of a custom rules file can be found at
        https://github.com/edx/bok-choy/tree/master/tests/a11y_custom_rules.js
        """
        custom_file = custom_ruleset_file or os.environ.get(
            "BOKCHOY_A11Y_CUSTOM_RULES_FILE"
        )

        if not custom_file:
            return

        with open(custom_file, "r") as additional_rules:
            custom_rules = additional_rules.read()

        if "var customRules" not in custom_rules:
            raise A11yAuditConfigError(
                "Custom rules file must include \"var customRules\""
            )

        self.custom_rules = custom_rules


class AxeCoreAudit(A11yAudit):
    """
    Use Deque Labs' axe-core engine to audit a page for accessibility issues.

    Related documentation:

        https://github.com/dequelabs/axe-core/blob/master/doc/API.md
    """

    @property
    def default_config(self):
        """
        Returns an instance of AxeCoreAuditConfig.
        """
        return AxeCoreAuditConfig()

    @staticmethod
    def _check_rules(browser, rules_js, config):
        """
        Run an accessibility audit on the page using the axe-core ruleset.

        Args:
            browser: a browser instance.
            rules_js: the ruleset JavaScript as a string.
            config: an AxsAuditConfig instance.

        Returns:
            A list of violations.

        Related documentation:

            https://github.com/dequelabs/axe-core/blob/master/doc/API.md#results-object

        __Caution__: You probably don't really want to call this method
        directly! It will be used by `AxeCoreAudit.do_audit`.
        """
        audit_run_script = dedent("""
            {rules_js}
            {custom_rules}
            axe.configure(customRules);
            var updatedResults = function(r) {{
                window.a11yAuditResults = JSON.stringify(r);
                window.console.log(window.a11yAuditResults);
            }}
            axe.a11yCheck({context}, {options}, updatedResults);
        """).format(
            rules_js=rules_js,
            custom_rules=config.custom_rules,
            context=config.context,
            options=config.rules
        )

        audit_results_script = dedent("""
            window.console.log(window.a11yAuditResults);
            return window.a11yAuditResults;
        """)

        browser.execute_script(audit_run_script)

        def audit_results_check_func():
            """
            A method to check that the audit has completed.

            Returns:

                (True, results) if the results are available.
                (False, None) if the results aren't available.
            """

            unicode_results = browser.execute_script(audit_results_script)

            try:
                results = json.loads(unicode_results)
            except (TypeError, ValueError):
                results = None

            if results:
                return True, results
            else:
                return False, None

        result = Promise(
            audit_results_check_func,
            "Timed out waiting for a11y audit results.",
            timeout=5,
        ).fulfill()

        # audit_results is report of accessibility violations for that session
        # Note that this ruleset doesn't have distinct error/warning levels.
        audit_results = result.get('violations')
        return audit_results

    @staticmethod
    def get_errors(audit_results):
        """
        Args:

            audit_results: results of `AxeCoreAudit.do_audit()`.

        Returns:

            A dictionary with keys "errors" and "total".
        """
        errors = {"errors": [], "total": 0}
        if audit_results:
            errors["errors"].extend(audit_results)
            for i in audit_results:
                for _node in i["nodes"]:
                    errors["total"] += 1
        return errors

    @staticmethod
    def format_errors(errors):
        """
        Args:

            errors: results of `AxeCoreAudit.get_errors()`.

        Returns: The errors as a formatted string.
        """

        def _get_message(node):
            """
            Get the message to display in the error output.
            """
            messages = set()

            try:
                messages.update([node['message']])
            except KeyError:
                pass

            for check_group in ['any', 'all', 'none']:
                try:
                    for check in node[check_group]:
                        messages.update([check.get('message')])
                except KeyError:
                    pass

            messages = messages.difference([''])
            return '; '.join(messages)

        lines = []
        for error_type in errors:
            lines.append("Severity: {}".format(error_type.get("impact")))
            lines.append("Rule ID: {}".format(error_type.get("id")))
            lines.append("Help URL: {}\n".format(error_type.get('helpUrl')))

            for node in error_type['nodes']:
                msg = "Message: {}".format(_get_message(node))
                html = "Html: {}".format(node.get('html').encode('utf-8'))
                target = "Target: {}".format(node.get('target'))
                fill_opts = {
                    'width': 100,
                    'initial_indent': '\t',
                    'subsequent_indent': '\t\t',
                }
                lines.append(fill(msg, **fill_opts))
                lines.append(fill(html, **fill_opts))
                lines.append(fill(target, **fill_opts))
                lines.append('\n')

        return '\n'.join(lines)

    @staticmethod
    def report_errors(audit, url):
        """
        Args:

            audit: results of `AxeCoreAudit.do_audit()`.
            url: the url of the page being audited.

        Raises: `AccessibilityError`
        """
        errors = AxeCoreAudit.get_errors(audit)
        if errors["total"] > 0:
            msg = "URL '{}' has {} errors:\n\n{}".format(
                url,
                errors["total"],
                AxeCoreAudit.format_errors(errors["errors"])
            )
            raise AccessibilityError(msg)



"""
Base implementation of the Page Object pattern.
See https://code.google.com/p/selenium/wiki/PageObjects
"""
from abc import ABCMeta, abstractmethod, abstractproperty
from collections import defaultdict
from functools import wraps
from contextlib import contextmanager
import logging
import os
import socket
import urlparse
import re
from textwrap import dedent
from lazy import lazy

from selenium.common.exceptions import WebDriverException

from .query import BrowserQuery
from .promise import Promise, EmptyPromise, BrokenPromise
from .a11y import AxeCoreAudit, AxsAudit


# String that can be used to test for XSS vulnerabilities.
# Taken from https://www.owasp.org/index.php/XSS_Filter_Evasion_Cheat_Sheet#XSS_Locator.
XSS_INJECTION = "'';!--\"<XSS>=&{()}"

# When the injected string appears within an attribute (for instance, value of an input tag,
# or alt of an img tag), if it is properly escaped this is the format we will see from
# document.documentElement.innerHTML. To avoid false positives, we need to allow this
# specific string, which hopefully is unique/odd enough that it would never appear accidentally.
EXPECTED_ATTRIBUTE_FORMAT = re.compile(r'\'\';!--&quot;<xss>=&amp;{\(\)}')

XSS_HTML = "<xss"


class WrongPageError(Exception):
    """
    The page object reports that we're on the wrong page!
    """
    pass


class PageLoadError(Exception):
    """
    An error occurred while loading the page.
    """
    pass


class XSSExposureError(Exception):
    """
    An XSS issue has been found on the current page.
    """
    pass


def unguarded(method):
    """
    Mark a PageObject method as unguarded.

    Unguarded methods don't verify that the PageObject is
    on the current browser page before they execute

    Args:
        method (callable): The method to decorate.

    Returns:
        Decorated method
    """
    method._unguarded = True  # pylint: disable=protected-access
    return method


def pre_verify(method):
    """
    Decorator that calls self._verify_page() before executing the decorated method

    Args:
        method (callable): The method to decorate.

    Returns:
        Decorated method
    """
    @wraps(method)
    def wrapper(self, *args, **kwargs):  # pylint: disable=missing-docstring
        self._verify_page()  # pylint: disable=protected-access
        return method(self, *args, **kwargs)
    return wrapper


class _PageObjectMetaclass(ABCMeta):
    """
    Decorates any callable attributes of the class
    so that they call self._verify_page() before executing.

    Excludes any methods marked as unguarded with the @unguarded
    decorator, any methods starting with _, or in the list ALWAYS_UNGUARDED.
    """
    ALWAYS_UNGUARDED = ['url', 'is_browser_on_page']

    def __new__(mcs, cls_name, cls_bases, cls_attrs):
        for name, attr in cls_attrs.items():
            # Skip methods marked as unguarded
            if getattr(attr, '_unguarded', False) or name in mcs.ALWAYS_UNGUARDED:
                continue

            # Skip private methods
            if name.startswith('_'):
                continue

            # Skip class attributes that are classes themselves
            if isinstance(attr, type):
                continue

            is_property = isinstance(attr, property)

            # Skip non-callable attributes
            if not (callable(attr) or is_property):
                continue

            if is_property:
                # For properties, wrap each of the sub-methods separately
                property_methods = defaultdict(None)
                for fn_name in ('fdel', 'fset', 'fget'):
                    prop_fn = getattr(cls_attrs[name], fn_name, None)
                    if prop_fn is not None:
                        # Check for unguarded properties
                        if getattr(prop_fn, '_unguarded', False):
                            property_methods[fn_name] = prop_fn
                        else:
                            property_methods[fn_name] = pre_verify(prop_fn)
                cls_attrs[name] = property(**property_methods)
            else:
                cls_attrs[name] = pre_verify(attr)

        return super(_PageObjectMetaclass, mcs).__new__(mcs, cls_name, cls_bases, cls_attrs)


class PageObject(object):
    """
    Encapsulates user interactions with a specific part
    of a web application.

    The most important thing is this:
    Page objects encapsulate Selenium.

    If you find yourself writing CSS selectors in tests,
    manipulating forms, or otherwise interacting directly
    with the web UI, stop!

    Instead, put these in a :class:`PageObject` subclass :)

    PageObjects do their best to verify that they are only
    used when the browser is on a page containing the object.
    To do this, they will call :meth:`is_browser_on_page` before executing
    any of their methods, and raise a :class:`WrongPageError` if the
    browser isn't on the correct page.

    Generally, this is the right behavior. However, at times it
    will be useful to not verify the page before executing a method.
    In those cases, the method can be marked with the :func:`unguarded`
    decorator. Additionally, private methods (those beginning with `_`)
    are always unguarded.

    Class or instance properties are never guarded. However, methods
    marked with the :func:`property` are candidates for being guarded.
    To make them unguarded, you must mark the getter, setter, and deleter
    as :func:`unguarded` separately, and those decorators must be applied before
    the :func:`property` decorator.

    Correct::

        @property
        @unguarded
        def foo(self):
            return self._foo

    Incorrect::

        @unguarded
        @property
        def foo(self):
            return self._foo
    """

    __metaclass__ = _PageObjectMetaclass

    def __init__(self, browser, *args, **kwargs):
        """
        Initialize the page object to use the specified browser instance.

        Args:
            browser (selenium.webdriver): The Selenium-controlled browser.

        Returns:
            PageObject
        """
        super(PageObject, self).__init__(*args, **kwargs)
        self.browser = browser
        a11y_flag = os.environ.get('VERIFY_ACCESSIBILITY', 'False')
        self.verify_accessibility = a11y_flag.lower() == 'true'
        xss_flag = os.environ.get('VERIFY_XSS', 'False')
        self.verify_xss = xss_flag.lower() == 'true'

    @lazy
    def a11y_audit(self):
        """
        Initializes the a11y_audit attribute.
        """
        rulesets = {
            "axe_core": AxeCoreAudit,
            "google_axs": AxsAudit,
        }

        ruleset = rulesets[
            os.environ.get("BOKCHOY_A11Y_RULESET", 'axe_core')]

        return ruleset(self.browser, self.url)

    @abstractmethod
    def is_browser_on_page(self):
        """
        Check that we are on the right page in the browser.
        The specific check will vary from page to page,
        but usually this amounts to checking the:

            1) browser URL
            2) page title
            3) page headings

        Returns:
            A `bool` indicating whether the browser is on the correct page.
        """
        return False

    @abstractproperty
    def url(self):
        """
        Return the URL of the page.  This may be dynamic,
        determined by configuration options passed to the
        page object's constructor.

        Some pages may not be directly accessible:
        perhaps the page object represents a "navigation"
        component that occurs on multiple pages.
        If this is the case, subclasses can return `None`
        to indicate that you can't directly visit the page object.
        """
        return None

    @unguarded
    def warning(self, msg):
        """
        Subclasses call this to indicate that something unexpected
        occurred while interacting with the page.

        Page objects themselves should never make assertions or
        raise exceptions, but they can issue warnings to make
        tests easier to debug.

        Args:
            msg (str): The message to log as a warning.

        Returns:
            None
        """
        log = logging.getLogger(self.__class__.__name__)
        log.warning(msg)

    @unguarded
    def visit(self):
        """
        Open the page containing this page object in the browser.

        Some page objects may not provide a URL, in which case
        a `NotImplementedError` will be raised.

        Raises:
            PageLoadError: The page did not load successfully.
            NotImplementedError: The page object does not provide a URL to visit.

        Returns:
            PageObject
        """
        if self.url is None:
            raise NotImplementedError("Page {} does not provide a URL to visit.".format(self))

        # Validate the URL
        if not self.validate_url(self.url):
            raise PageLoadError("Invalid URL: '{}'".format(self.url))

        # Visit the URL
        try:
            self.browser.get(self.url)
        except (WebDriverException, socket.gaierror):
            raise PageLoadError("Could not load page '{!r}' at URL '{}'".format(
                self, self.url
            ))

        # Give the browser enough time to get to the page, then return the page object
        # so that the caller can chain the call with an action:
        # Example: FooPage.visit().do_something()
        #
        # A BrokenPromise will be raised if the page object's is_browser_on_page method
        # does not return True before timing out.
        try:
            return self.wait_for_page()
        except BrokenPromise:
            raise PageLoadError("Timed out waiting to load page '{!r}' at URL '{}'".format(
                self, self.url
            ))

    @classmethod
    @unguarded
    def validate_url(cls, url):
        """
        Return a boolean indicating whether the URL has a protocol and hostname.
        If a port is specified, ensure it is an integer.

        Arguments:
            url (str): The URL to check.

        Returns:
            Boolean indicating whether the URL has a protocol and hostname.
        """
        result = urlparse.urlsplit(url)

        # Check that we have a protocol and hostname
        if not result.scheme or not result.netloc:
            return False

        # Check that the port is an integer
        try:
            if result.port is not None:
                int(result.port)
            elif result.netloc.endswith(':'):
                # Valid URLs do not end with colons.
                return False
        except ValueError:
            return False
        else:
            return True

    def _verify_page(self):
        """
        Ask the page object if we're on the right page;
        if not, raise a `WrongPageError`.
        """
        if not self.is_browser_on_page():
            msg = "Not on the correct page to use '{!r}' at URL '{}'".format(
                self, self.url
            )
            raise WrongPageError(msg)

    def _verify_xss_exposure(self):
        """
        Verify that there are no obvious XSS exposures on the page (based on test authors
        including XSS_INJECTION in content rendered on the page).

        If an xss issue is found, raise a 'XSSExposureError'.
        """
        # Use innerHTML to get dynamically injected HTML as well as server-side HTML.
        html_source = self.browser.execute_script(
            "return document.documentElement.innerHTML.toLowerCase()"
        )

        # Check taken from https://www.owasp.org/index.php/XSS_Filter_Evasion_Cheat_Sheet#XSS_Locator.
        all_hits_count = html_source.count(XSS_HTML)
        if all_hits_count > 0:
            safe_hits_count = len(EXPECTED_ATTRIBUTE_FORMAT.findall(html_source))
            if all_hits_count > safe_hits_count:
                potential_hits = re.findall('<[^<]+<xss', html_source)
                raise XSSExposureError(
                    "{} XSS issue(s) found on page. Potential places are {}".format(
                        all_hits_count - safe_hits_count, potential_hits
                    )
                )

    @unguarded
    def wait_for_page(self, timeout=30):
        """
        Block until the page loads, then returns the page.
        Useful for ensuring that we navigate successfully to a particular page.

        Keyword Args:
            timeout (int): The number of seconds to wait for the page before timing out with an exception.

        Raises:
            BrokenPromise: The timeout is exceeded without the page loading successfully.
        """
        def _is_document_ready():
            """
            Check the loading state of the document to ensure the document and all sub-resources
            have finished loading (the document load event has been fired.)
            """
            return self.browser.execute_script(
                "return document.readyState=='complete'")

        EmptyPromise(
            _is_document_ready,
            "The document and all sub-resources have finished loading.",
            timeout=timeout
        ).fulfill()

        result = Promise(
            lambda: (self.is_browser_on_page(), self), "loaded page {!r}".format(self),
            timeout=timeout
        ).fulfill()

        if self.verify_accessibility:
            self.a11y_audit.check_for_accessibility_errors()  # pylint: disable=no-member

        return result

    @unguarded
    def q(self, **kwargs):  # pylint: disable=invalid-name
        """
        Construct a query on the browser.

        Example usages:

        .. code:: python

            self.q(css="div.foo").first.click()
            self.q(xpath="/foo/bar").text

        Keyword Args:
            css: A CSS selector.
            xpath: An XPath selector.

        Returns:
            BrowserQuery
        """
        if self.verify_xss:
            self._verify_xss_exposure()
        return BrowserQuery(self.browser, **kwargs)

    @contextmanager
    def handle_alert(self, confirm=True):
        """
        Context manager that ensures alerts are dismissed.

        Example usage:

        .. code:: python

            with self.handle_alert():
                self.q(css='input.submit-button').first.click()

        Keyword Args:
            confirm (bool): Whether to confirm or cancel the alert.

        Returns:
            None
        """

        # Before executing the `with` block, stub the confirm/alert functions
        script = dedent("""
            window.confirm = function() {{ return {0}; }};
            window.alert = function() {{ return; }};
        """.format("true" if confirm else "false")).strip()
        self.browser.execute_script(script)

        # Execute the `with` block
        yield

    @unguarded
    def wait_for_ajax(self, timeout=30):
        """
        Wait for jQuery to be loaded and for all ajax requests to finish. Note
        that we have to wait for jQuery to load first because it is used to
        check that ajax requests are complete.

        Important: If you have an ajax requests that results in a page reload,
        you will need to use wait_for_page or some other method to confirm that
        the page has finished reloading after wait_for_ajax has returned.

        Example usage:

        .. code:: python

            self.q(css='input#email').fill("foo")
            self.wait_for_ajax()

        Keyword Args:
            timeout (int): The number of seconds to wait before timing out with
            a BrokenPromise exception.

        Returns:
            None

        Raises:
            BrokenPromise: The timeout is exceeded before (1) jQuery is defined
            and (2) all ajax requests are completed.
        """

        def _is_ajax_finished():
            """
            Check if all the ajax calls on the current page have completed.
            """
            # Wait for jQuery to be defined first, so that jQuery.active
            # doesn't raise an error that 'jQuery is not defined'.  We have
            # seen this as a flaky pattern possibly related to pages reloading
            # while wait_for_ajax is being called.
            return self.browser.execute_script(
                "return typeof(jQuery)!='undefined' && jQuery.active==0")

        EmptyPromise(
            _is_ajax_finished,
            "Finished waiting for ajax requests.",
            timeout=timeout
        ).fulfill()

    @unguarded
    def wait_for(self, promise_check_func, description, result=False, timeout=60):
        """
        Calls the method provided as an argument until the Promise satisfied or BrokenPromise

        Arguments:
            promise_check_func (callable):
                * If `result` is False Then
                    Function that accepts no arguments and returns a boolean indicating whether the promise is fulfilled
                * If `result` is True Then
                    Function that accepts no arguments and returns a `(is_satisfied, result)` tuple,
                    where `is_satisfied` is a boolean indicating whether the promise was satisfied, and `result`
                    is a value to return from the fulfilled `Promise`
            description (str): Description of the Promise, used in log messages
            result (bool): Indicates whether we need result
            timeout (float): Maximum number of seconds to wait for the Promise to be satisfied before timing out

        Raises:
            BrokenPromise: the `Promise` was not satisfied

        """
        if result:
            return Promise(promise_check_func, description, timeout=timeout).fulfill()
        else:
            return EmptyPromise(promise_check_func, description, timeout=timeout).fulfill()

    @unguarded
    def wait_for_element_presence(self, element_selector, description, timeout=60):
        """
        Waits for element specified by `element_selector` to be present in DOM.

        Example usage:

        .. code:: python

            self.wait_for_element_presence('.submit', 'Submit Button is Present')

        Arguments:
            element_selector (str): css selector of the element.
            description (str): Description of the Promise, used in log messages.
            timeout (float): Maximum number of seconds to wait for the Promise to be satisfied before timing out

        """
        self.wait_for(lambda: self.q(css=element_selector).present, description=description, timeout=timeout)

    @unguarded
    def wait_for_element_absence(self, element_selector, description, timeout=60):
        """
        Waits for element specified by `element_selector` until it disappears from DOM.

        Example usage:

        .. code:: python

            self.wait_for_element_absence('.submit', 'Submit Button is not Present')

        Arguments:
            element_selector (str): css selector of the element.
            description (str): Description of the Promise, used in log messages.
            timeout (float): Maximum number of seconds to wait for the Promise to be satisfied before timing out

        """
        self.wait_for(lambda: not self.q(css=element_selector).present, description=description, timeout=timeout)

    @unguarded
    def wait_for_element_visibility(self, element_selector, description, timeout=60):
        """
        Waits for element specified by `element_selector` until it is displayed on web page.

        Example usage:

        .. code:: python

            self.wait_for_element_visibility('.submit', 'Submit Button is Visible')

        Arguments:
            element_selector (str): css selector of the element.
            description (str): Description of the Promise, used in log messages.
            timeout (float): Maximum number of seconds to wait for the Promise to be satisfied before timing out

        """
        self.wait_for(lambda: self.q(css=element_selector).visible, description=description, timeout=timeout)

    @unguarded
    def wait_for_element_invisibility(self, element_selector, description, timeout=60):
        """
        Waits for element specified by `element_selector` until it disappears from the web page.

        Example usage:

        .. code:: python

            self.wait_for_element_invisibility('.submit', 'Submit Button Disappeared')

        Arguments:
            element_selector (str): css selector of the element.
            description (str): Description of the Promise, used in log messages.
            timeout (float): Maximum number of seconds to wait for the Promise to be satisfied before timing out

        """
        self.wait_for(lambda: self.q(css=element_selector).invisible, description=description, timeout=timeout)

    @unguarded
    def scroll_to_element(self, element_selector, timeout=60):
        """
        Scrolls the browser such that the element specified appears at the top. Before scrolling, waits for
        the element to be present.

        Example usage:

        .. code:: python

            self.scroll_to_element('.far-down', 'Scroll to far-down')

        Arguments:
            element_selector (str): css selector of the element.
            timeout (float): Maximum number of seconds to wait for the element to be present on the
                page before timing out.

        Raises: BrokenPromise if the element does not exist (and therefore scrolling to it is not possible)

        """
        # Ensure element exists
        msg = "Element '{element}' is present".format(element=element_selector)
        self.wait_for(lambda: self.q(css=element_selector).present, msg, timeout=timeout)

        # Obtain coordinates and use those for JavaScript call
        loc = self.q(css=element_selector).first.results[0].location
        self.browser.execute_script("window.scrollTo({x},{y})".format(x=loc['x'], y=loc['y']))



"""
Interface for running accessibility audits on a PageObject.
"""
import os
from abc import abstractmethod, abstractproperty, ABCMeta


class AccessibilityError(Exception):
    """
    The page violates one or more accessibility rules.
    """
    pass


class A11yAuditConfigError(Exception):
    """
    An error in A11yAuditConfig.
    """
    pass


class A11yAuditConfig(object):
    """
    The `A11yAuditConfig` object defines the options available in an
    accessibility ruleset.
    """
    __metaclass__ = ABCMeta

    def __init__(self, *args, **kwargs):
        super(A11yAuditConfig, self).__init__(*args, **kwargs)
        self.rules_file = None

    def set_rules_file(self, path=None):
        """
        Sets `self.rules_file` to the passed file.

        Args:
            A filepath where the JavaScript for the ruleset can be found.

        This is intended to be used in the case of using an extended or
        modified version of the ruleset. The interface and response
        format are expected to be unmodified.
        """
        if path:
            self.rules_file = os.path.abspath(path)

    @abstractmethod
    def set_rules(self, rules):
        """
        Overrides the default rules to be run.

        Raises:

            `NotImplementedError` if this isn't overwritten in the ruleset
                specific implementation.
        """
        raise NotImplementedError(
            "The ability to specify rules has not been implemented."
        )

    @abstractmethod
    def set_scope(self, include=None, exclude=None):
        """
        Overrides the default scope (part of the DOM) to inspect.

        Raises:

            `NotImplementedError` if this isn't overwritten in the ruleset
                specific implementation.
        """
        raise NotImplementedError(
            "The ability to specify scope has not been implemented."
        )

    @abstractmethod
    def customize_ruleset(self, custom_ruleset_file=None):
        """
        Allows customization of the ruleset. (e.g. adding custom rules,
        extending the implementation of an existing rule.)

        Raises:

            `NotImplementedError` if this isn't overwritten in the ruleset
                specific implementation.
        """
        raise NotImplementedError(
            "The ability to customize the ruleset has not been implemented."
        )


class A11yAudit(object):
    """
    Allows auditing of a page for accessibility issues.

    The ruleset to use can be specified by the environment variable
    `BOKCHOY_A11Y_RULESET`. Currently, there are two ruleset implemented:

    `axe_core`:

        * Ruleset class: AxeCoreAudit
        * Ruleset config: AxeCoreAuditConfig
        * This is default ruleset.

    `google_axs`:

        * Ruleset class: AxsAudit
        * Ruleset config: AxsAuditConfig
    """
    __metaclass__ = ABCMeta

    def __init__(self, browser, url, config=None, *args, **kwargs):
        """
        Sets ruleset to be used.

        Args:
            browser: A browser instance
            url: URL of the page to test
            config: (optional) A11yAuditConfig or subclass of A11yAuditConfig
        """
        super(A11yAudit, self).__init__(*args, **kwargs)
        self.url = url
        self.browser = browser
        self.config = config or self.default_config

    def _get_rules_js(self):
        """
        Checks that the rules file for the enabled ruleset exists
        and returns its contents as string.

        Raises: `RuntimeError` if the file isn't found.
        """
        if not os.path.isfile(self.config.rules_file):
            msg = 'Could not find the accessibility tools JS file: {}'.format(
                self.config.rules_file)
            raise RuntimeError(msg)

        else:
            with open(self.config.rules_file, "r") as rules_file:
                return rules_file.read()

    def do_audit(self):
        """
        Audit the page for accessibility problems using the enabled ruleset.

        Returns:
            A list (one for each browser session) of results returned from
            the audit. See documentation of `_check_rules` in the enabled
            ruleset for the format of each result.
        """
        rules_js = self._get_rules_js()
        audit_results = self._check_rules(
            self.browser, rules_js, self.config)
        return audit_results

    def check_for_accessibility_errors(self):
        """
        Run an accessibility audit, parse the results, and raise a single
        exception if there are violations.

        Note that an exception is only raised on errors, not on warnings.

        Returns:

            None

        Raises:

            AccessibilityError
        """
        audit_results = self.do_audit()
        if audit_results:
            self.report_errors(audit_results, self.url)

    @abstractproperty
    def default_config(self):
        """
        Return an instance of a subclass of A11yAuditConfig.
        """
        raise NotImplementedError("default_config has not been implemented")

    @staticmethod
    @abstractmethod
    def _check_rules(browser, rules_js, config):
        """
        Run an accessibility audit on the page using the implemented ruleset.

        Args:

            browser: a browser instance.
            rules_js: the ruleset JavaScript as a string.
            config: an AxsAuditConfig instance.

        Returns:

            A list of violations.

        Raises:

            `NotImplementedError` if this isn't overwritten in the ruleset
                specific implementation.
        """
        raise NotImplementedError("_check_rules has not been implemented")

    @staticmethod
    @abstractmethod
    def report_errors(audit, url):
        """
        Args:

            audit: results of an accessibility audit.
            url: the url of the page being audited.

        Raises:

            `AccessibilityError` if errors are found in the audit.
            `NotImplementedError` if this isn't overwritten in the ruleset
                specific implementation.
        """
        raise NotImplementedError("report_errors has not been implemented")



"""
Variation on the "promise" design pattern.
Promises make it easier to handle asynchronous operations correctly.
"""
import time
import logging


LOGGER = logging.getLogger(__name__)


class BrokenPromise(Exception):
    """
    The promise was not satisfied within the time constraints.
    """

    def __init__(self, promise):
        """
        Configure the broken promise error.

        Args:
            promise (Promise): The promise that was not satisfied.
        """
        super(BrokenPromise, self).__init__()
        self._promise = promise

    def __str__(self):
        return "Promise not satisfied: {0}".format(self._promise)


class Promise(object):
    """
    Check that an asynchronous action completed, blocking until it does
    or timeout / try limits are reached.
    """

    def __init__(self, check_func, description, try_limit=None, try_interval=0.5, timeout=30):
        """
        Configure the `Promise`.

        The `Promise` will poll `check_func()` until either:
            * The promise is satisfied
            * The promise runs out of tries (checks more than `try_limit` times)
            * The promise runs out of time (takes longer than `timeout` seconds)

        In the second two cases, the promise is "broken" and an exception will be raised.
        `description` is a string that will be included in the exception to make debugging easier.

        Example:

        .. code:: python

            # Dummy check function that indicates the promise is always satisfied
            check_func = lambda: (True, "Hello world!")

            # Check up to 5 times if the operation has completed
            result = Promise(check_func, "Operation has completed", try_limit=5).fulfill()

        Args:
            check_func (callable): A function that accepts no arguments and returns a `(is_satisfied, result)` tuple,
                where `is_satisfied` is a boolean indiating whether the promise was satisfied, and `result`
                is a value to return from the fulfilled `Promise`.

            description (str): Description of the `Promise`, used in log messages.

        Keyword Args:
            try_limit (int or None): Number of attempts to make to satisfy the `Promise`.
                Can be `None` to disable the limit.
            try_interval (float): Number of seconds to wait between attempts.
            timeout (float): Maximum number of seconds to wait for the `Promise` to be satisfied before timing out.

        Returns:
            Promise
        """
        self._check_func = check_func
        self._description = description
        self._try_limit = try_limit
        self._try_interval = try_interval
        self._timeout = timeout
        self._num_tries = 0

    def fulfill(self):
        """
        Evaluate the promise and return the result.

        Returns:
             The result of the `Promise` (second return value from the `check_func`)

        Raises:
            BrokenPromise: the `Promise` was not satisfied within the time or attempt limits.
        """
        is_fulfilled, result = self._check_fulfilled()

        if is_fulfilled:
            return result
        else:
            raise BrokenPromise(self)

    def __str__(self):
        return str(self._description)

    def _check_fulfilled(self):
        """
        Return tuple `(is_fulfilled, result)` where
        `is_fulfilled` is a boolean indicating whether the promise has been fulfilled
        and `result` is the value to pass to the `with` block.
        """
        is_fulfilled = False
        result = None
        start_time = time.time()

        # Check whether the promise has been fulfilled until we run out of time or attempts
        while self._has_time_left(start_time) and self._has_more_tries():

            # Keep track of how many attempts we've made so far
            self._num_tries += 1

            is_fulfilled, result = self._check_func()

            # If the promise is satisfied, then continue execution
            if is_fulfilled:
                break

            # Delay between checks
            time.sleep(self._try_interval)

        return is_fulfilled, result

    def _has_time_left(self, start_time):
        """
        Return True if the elapsed time is less than the timeout.
        """
        return time.time() - start_time < self._timeout

    def _has_more_tries(self):
        """
        Return True if the promise has additional tries.
        If `_try_limit` is `None`, always return True.
        """
        if self._try_limit is None:
            return True
        else:
            return self._num_tries < self._try_limit


class EmptyPromise(Promise):
    """
    A promise that has no result value.
    """

    def __init__(self, check_func, description, **kwargs):
        """
        Configure the promise.

        Unlike a regular `Promise`, the `check_func()` does NOT return a tuple
        with a result value.  That's why the promise is "empty" -- you don't get anything back.

        Example usage:

        .. code:: python

            # This will block until `is_done` returns `True` or we reach the timeout limit.
            EmptyPromise(lambda: is_done('test'), "Test operation is done").fulfill()

        Args:
            check_func (callable): Function that accepts no arguments and
                returns a boolean indicating whether the promise is fulfilled.
            description (str): Description of the Promise, used in log messages.

        Returns:
            EmptyPromise
        """
        full_check_func = lambda: (check_func(), None)
        super(EmptyPromise, self).__init__(full_check_func, description, **kwargs)



#!/usr/bin/env python

import codecs
from setuptools import setup

VERSION = '0.5.4'
DESCRIPTION = 'UI-level acceptance test framework'
REQUIREMENTS = (
    'lazy',
    'needle',
    'selenium',
)

with codecs.open('README.rst', 'r', 'utf-8') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='bok_choy',
    version=VERSION,
    author='edX',
    author_email='oscm@edx.org',
    url='http://github.com/edx/bok-choy',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    license='Apache 2.0',
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: Apache Software License',
                 'Operating System :: OS Independent',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2',
                 'Programming Language :: Python :: 2.7',
                 'Topic :: Software Development :: Testing',
                 'Topic :: Software Development :: Quality Assurance'],
    packages=['bok_choy', 'bok_choy/a11y'],
    package_data={'bok_choy': ['vendor/google/*.*', 'vendor/axe-core/*.*']},
    install_requires=REQUIREMENTS,
)
